/**
 * \file mod_example2.c
 * \author Christian Kruse, <ckruse@wwwtech.de>
 *
 * A module executing CGI-Requests depending on the load of the server
 */

/* {{{ Apache license
 * ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */
/* }}} */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
#include "http_connection.h"

#include "apr_strings.h"

#include <stdio.h>

/**
 * Module config structure
 */
typedef struct s_mod_loadavg_cfg {
  #define MOD_LOADAVG_CFG_SERVER 1
  #define MOD_LOADAVG_CFG_DIR    2
  #define MOD_LOADAVG_CFG_COMBO  3     /* Shouldn't ever happen. */
  int cmode;

  int loadavg_on;
  int loadavg_on_set;

  double loadavg_1;
  int loadavg_1_set;

  double loadavg_5;
  int loadavg_5_set;

  double loadavg_15;
  int loadavg_15_set;
} t_mod_loadavg_cfg;

/*
 * Declare ourselves so the configuration routines can find and know us.
 * We'll fill it in at the end of the module.
 */
module AP_MODULE_DECLARE_DATA loadavg_module;

static const char *handle_loadavg_on(cmd_parms *cmd,void *mconfig,int onoff) {
  t_mod_loadavg_cfg *cfg = (t_mod_loadavg_cfg *)mconfig;

  cfg->loadavg_on = onoff;
  cfg->loadavg_on_set = 1;

  return NULL;
}

static const char *handle_loadavg_1(cmd_parms *cmd, void *mconfig,char *val) {
  t_mod_loadavg_cfg *cfg = (t_mod_loadavg_cfg *)mconfig;
  double dval;
  char *ptr = NULL;

  dval = strtod(val,&ptr);
  if(dval == .0 && ptr == val) return "Valid values for the Loadavg1 directive is a floating point number";

  cfg->loadavg_1 = dval;
  cfg->loadavg_1_set = 1;

  return NULL;
}

static const char *handle_loadavg_5(cmd_parms *cmd, void *mconfig,char *val) {
  t_mod_loadavg_cfg *cfg = (t_mod_loadavg_cfg *)mconfig;
  double dval;
  char *ptr = NULL;

  dval = strtod(val,&ptr);
  if(dval == (double)0 && ptr == val) return "Valid values for the Loadavg5 directive is a floating point number";

  cfg->loadavg_5 = dval;
  cfg->loadavg_5_set = 1;

  return NULL;
}

static const char *handle_loadavg_15(cmd_parms *cmd, void *mconfig,char *val) {
  t_mod_loadavg_cfg *cfg = (t_mod_loadavg_cfg *)mconfig;
  double dval;
  char *ptr = NULL;

  dval = strtod(val,&ptr);
  if(dval == (double)0 && ptr == val) return "Valid values for the Loadavg15 directive is a floating point number";

  cfg->loadavg_15 = dval;
  cfg->loadavg_15_set = 1;

  return NULL;
}



void mod_loadavg_default_values(t_mod_loadavg_cfg *cfg) {
  cfg->cmode      = 0;
  cfg->loadavg_on = 0;

  cfg->loadavg_1  = 4.0;
  cfg->loadavg_5  = (double)0;
  cfg->loadavg_15 = (double)0;

  cfg->loadavg_on_set = 0;
  cfg->loadavg_1_set  = 0;
  cfg->loadavg_5_set  = 0;
  cfg->loadavg_15_set = 0;
}

static void *mod_loadavg_create_dir_config(apr_pool_t *p, char *dirspec) {
  t_mod_loadavg_cfg *cfg = apr_palloc(p, sizeof(*cfg));

  /* fill with default values */
  mod_loadavg_default_values(cfg);

  cfg->cmode = MOD_LOADAVG_CFG_DIR;

  return (void *)cfg;
}

static void *mod_loadavg_merge_dir_config(apr_pool_t *p, void *parent_conf,void *newloc_conf) {
  t_mod_loadavg_cfg *merged_config = apr_palloc(p, sizeof(*merged_config));
  t_mod_loadavg_cfg *par_conf      = (t_mod_loadavg_cfg *)parent_conf;
  t_mod_loadavg_cfg *new_conf      = (t_mod_loadavg_cfg *)newloc_conf;

  /* we will override server config by directory config */

  if(new_conf->loadavg_on_set) {
    merged_config->loadavg_on = new_conf->loadavg_on;
    merged_config->loadavg_on_set = 1;
  }
  else {
    merged_config->loadavg_on     = par_conf->loadavg_on;
    merged_config->loadavg_on_set = par_conf->loadavg_on_set;
  }

  if(new_conf->loadavg_1_set) {
    merged_config->loadavg_1     = new_conf->loadavg_1;
    merged_config->loadavg_1_set = 1;
  }
  else {
    merged_config->loadavg_1     = par_conf->loadavg_1;
    merged_config->loadavg_1_set = par_conf->loadavg_1_set;
  }

  if(new_conf->loadavg_5_set) {
    merged_config->loadavg_5     = new_conf->loadavg_5;
    merged_config->loadavg_5_set = 1;
  }
  else {
    merged_config->loadavg_5     = par_conf->loadavg_5;
    merged_config->loadavg_5_set = par_conf->loadavg_5_set;
  }

  if(new_conf->loadavg_15_set) {
    merged_config->loadavg_15     = new_conf->loadavg_15;
    merged_config->loadavg_15_set = 1;
  }
  else {
    merged_config->loadavg_15     = par_conf->loadavg_15;
    merged_config->loadavg_15_set = par_conf->loadavg_15_set;
  }

  return (void *)merged_config;
}

static void *mod_loadavg_create_server_config(apr_pool_t *p, server_rec *s) {
  t_mod_loadavg_cfg *cfg = apr_palloc(p, sizeof(*cfg));

  /* fill with default values */
  mod_loadavg_default_values(cfg);

  cfg->cmode = MOD_LOADAVG_CFG_SERVER;

  return (void *)cfg;
}

static void *mod_loadavg_merge_server_config(apr_pool_t *p, void *server1_conf,void *server2_conf) {
  return mod_loadavg_merge_dir_config(p,server1_conf,server2_conf);
}


static int mod_loadavg_type_checker(request_rec *r) {
  t_mod_loadavg_cfg *dir_cfg;
  double load_values[3] = { 0, 0, 0 };
  int processes;

  dir_cfg = (t_mod_loadavg_cfg *)ap_get_module_config(r->per_dir_config,&loadavg_module);

  ap_log_error(
    APLOG_MARK,
    APLOG_DEBUG,
    0,
    r->server,
    "called for request... mod_loadavg is %s!",apr_pstrdup(r->pool,dir_cfg->loadavg_on?"ON":"OFF")
  );

  if(dir_cfg->loadavg_on) {
    if(dir_cfg->loadavg_1_set || dir_cfg->loadavg_5_set || dir_cfg->loadavg_15_set) {
      if((processes = getloadavg(load_values,3)) == -1) {
        ap_log_error(APLOG_MARK,APLOG_NOTICE,0,NULL,"getloadavg: %s",strerror(errno));
        return DECLINED;
      }

      if(dir_cfg->loadavg_1_set) {
        if(dir_cfg->loadavg_1 < load_values[0]) {
          ap_log_error(APLOG_MARK,APLOG_NOTICE,0,NULL,"declined request from client %s for %s!",r->connection->remote_ip,r->filename);
          return HTTP_SERVICE_UNAVAILABLE;
        }
      }

      if(dir_cfg->loadavg_5_set) {
        if(dir_cfg->loadavg_5 < load_values[1]) {
          ap_log_error(APLOG_MARK,APLOG_NOTICE,0,NULL,"declined request from client %s for %s!",r->connection->remote_ip,r->filename);
          return HTTP_SERVICE_UNAVAILABLE;
        }
      }

      if(dir_cfg->loadavg_15_set) {
        if(dir_cfg->loadavg_15 < load_values[2]) {
          ap_log_error(APLOG_MARK,APLOG_NOTICE,0,NULL,"declined request from client %s for %s!",r->connection->remote_ip,r->filename);
          return HTTP_SERVICE_UNAVAILABLE;
        }
      }
    }
  }

  return DECLINED;
}


static int mod_loadavg_init_handler(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s) {
  ap_add_version_component(p, "mod_loadavg/0.1");
  return OK;
}


static void mod_loadavg_register_hooks(apr_pool_t *p) {
  ap_hook_post_config(mod_loadavg_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_type_checker(mod_loadavg_type_checker, NULL, NULL, APR_HOOK_FIRST);
}

static const command_rec mod_loadavg_cmds[] = {
  AP_INIT_FLAG("ModLoadavg",(void *)handle_loadavg_on,NULL,OR_OPTIONS,"set mod_loadavg on or off"),
  AP_INIT_TAKE1("Loadavg1",(void *)handle_loadavg_1,NULL,OR_OPTIONS,"set max value for the load average of the last minute"),
  AP_INIT_TAKE1("Loadavg5",(void *)handle_loadavg_5,NULL,OR_OPTIONS,"set max value for the load average of the last 5 minutes"),
  AP_INIT_TAKE1("Loadavg15",(void *)handle_loadavg_15,NULL,OR_OPTIONS,"set max value for the load average of the last 15 minutes"),
  {NULL}
};

module AP_MODULE_DECLARE_DATA loadavg_module = {
  STANDARD20_MODULE_STUFF,
  mod_loadavg_create_dir_config,    /* per-directory config creator */
  mod_loadavg_merge_dir_config,     /* dir config merger */
  mod_loadavg_create_server_config, /* server config creator */
  mod_loadavg_merge_server_config,  /* server config merger */
  mod_loadavg_cmds,                 /* command table */
  mod_loadavg_register_hooks        /* set up other request processing hooks */
};

/* eof */
